#!/usr/bin/env python

"""CCSDS TDM/OEM file input outputs

Orbit Data Messages
ODM     502.0-B-2   Blue Book (11/2009)
https://public.ccsds.org/Pubs/502x0b2c1e2.pdf

Tracking Data Message
TDM     503.0-B-2   Blue Book (06/2020)
https://public.ccsds.org/Pubs/503x0b2.pdf

NPM (Navigation Data Messages)
https://public.ccsds.org/Pubs/505x0b1.pdf

NPM Schemas covering TDM and OEM and more
https://sanaregistry.org/files/ndmxml_unqualified/*
https://sanaregistry.org/files/ndmxml_qualified/*


"""
import xmlschema
import numpy as np
import scipy.constants as consts
from astropy.time import TimeISO, Time
from ..version import __version__

from sorts.data import DATA


class epochType(TimeISO):
    """Convert astropy time to "yyyy-dddThh:mm:ss.d->dZ" CCSDS 2.0 epochType standard."""

    name = "CCSDS_epoch"  # Unique format name
    subfmts = (
        (
            "date_hms",
            "%Y-%jT%H:%M:%S",
            "{year:d}-{yday:03d}T{hour:02d}:{min:02d}:{sec:02d}",
        ),
    )


###############################################################################
# NDM SCHEMA
###############################################################################

# needs only to be unique - does not matter what it points to
# blue book examples for both tdm versions 1 and version 2 have this,
# (even though it points to what appears to be a version 1 specific resource)
# either way, that uri does not resolve

_SCHEMA_URI = "http://sanaregistry.org/r/ndmxml/ndmxml-1.0-master.xsd"


_SCHEMA = None


def get_schema():
    global _SCHEMA
    if _SCHEMA is None:
        xsd_path = DATA["ndmxml-2.0.0-master-2.0.xsd"]
        _SCHEMA = xmlschema.XMLSchema(xsd_path)
    return _SCHEMA


###############################################################################
# OEM
###############################################################################

_OEM_METADATA_FIELDS = [
    "COMMENT",
    "OBJECT_NAME",
    "OBJECT_ID",
    "CENTER_NAME",
    "REF_FRAME",
    "REF_FRAME_EPOCH",
    "TIME_SYSTEM",
    "START_TIME",
    "USEABLE_START_TIME",
    "USEABLE_STOP_TIME",
    "STOP_TIME",
    "INTERPOLATION",
    "INTERPOLATION_DEGREE",
]

###############################################################################
# OEM READ
###############################################################################


def read_xml_oem(xml):
    """
    read xml OEM file resource, parse and validate according to schema,
    returns python dictionary representing document
    """
    schema = get_schema()
    d = schema.decode(xml)
    return d


###############################################################################
# OEM WRITE
###############################################################################


def write_xml_oem(state_data, cov_data=None, meta=None, file=None):
    """
    creaate and validate xml for OEM message
    write xml file resource

    #TODO: finish this docstring

    assumes UTC input times

    It is assumed that the covariance matrix reference frame is the same as
    the state vector reference frame, i.e. `COV_REF_FRAME = meta['REF_FRAME']`.

    :param numpy.ndarray state_data: structured numpy array where each element
        in the array contains the data fields with the same name as those
        defined by the OEM XML for states.
    :param numpy.ndarray cov_data: structured numpy array where each element
        in the array contains the data fields with the same name as those
        defined by the OEM XML for covariance matricies.
    :param dict meta: .... Only exception is `DATA_COMMENT`, which is used as
        input for the `COMMENT` field for the data section rather then a
        `DATA_COMMENT` field in the meta data section.
    """
    if meta is None:
        meta = {}
    # originator
    originator = meta.get("ORIGINATOR", f"SORTS {__version__}")

    # creation data
    creation_date = meta.get("CREATION_DATE", Time.now()).CCSDS_epoch

    # tdm dictionary
    d = {
        "@id": "CCSDS_OEM_VERS",
        "@version": "2.0",
        "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "@xsi:noNamespaceSchemaLocation": _SCHEMA_URI,
        "header": {
            "COMMENT": "HEADER COMMENT",
            "CREATION_DATE": f"{creation_date}",
            "ORIGINATOR": f"{originator}",
        },
        "body": {"segment": []},
    }

    # ALTERNATIVES
    #
    # 1) - make one segment per observation or
    # 2) - one segment with multiple observations
    #
    # CHOOSE 2)
    segment = {"metadata": {}, "data": {}}
    d["body"]["segment"].append(segment)

    ###########################################################################
    # SEGMENT METADATA
    ###########################################################################

    _meta = segment["metadata"]

    # mandatory
    meta_defaults = dict(
        OBJECT_NAME="missing",
        OBJECT_ID="missing",
        CENTER_NAME="missing",
        REF_FRAME="missing",
        TIME_SYSTEM="UTC",
        START_TIME=Time.now().CCSDS_epoch,
        STOP_TIME=Time.now().CCSDS_epoch,
    )

    # add values from given meta - in correct order
    for field in _OEM_METADATA_FIELDS:
        if field in meta:
            _meta[field] = meta[field]
            # ensure correct CCSDS formatting of input astropy time object
            if field.endswith("_TIME"):
                _meta[field] = _meta[field].CCSDS_epoch
        else:
            if field in meta_defaults:
                _meta[field] = meta_defaults[field]

    ###########################################################################
    # SEGMENT DATA
    ###########################################################################

    _data = segment["data"]
    if "DATA_COMMENT" in meta:
        _data["COMMENT"] = meta["DATA_COMMENT"]

    _data["stateVector"] = []
    _data["covarianceMatrix"] = []

    # mock_vectors = [
    #     [Time.now().CCSDS_epoch, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9],
    #     [Time.now().CCSDS_epoch, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9],
    # ]

    for v in state_data:
        _dat = dict(
            EPOCH=Time(v["EPOCH"], format="datetime64", scale="utc").CCSDS_epoch,
            X=v["X"],
            Y=v["Y"],
            Z=v["Z"],
            X_DOT=v["X_DOT"],
            Y_DOT=v["Y_DOT"],
            Z_DOT=v["Z_DOT"],
        )
        # optional
        for key in ["X_DDOT", "Y_DDOT", "Z_DDOT"]:
            if key in state_data.dtype.names:
                _dat[key] = v[key]

        _data["stateVector"].append(_dat)

    if cov_data is not None:
        for v in cov_data:
            _dat_cov = dict()
            if "COMMENT" in cov_data.dtype.names:
                try:
                    _comment = v["COMMENT"].decode()
                except AttributeError:
                    _comment = v["COMMENT"]

                _dat_cov["COMMENT"] = _comment
                # remove empty comments
                if len(_dat_cov["COMMENT"]) == 0:
                    del _dat_cov["COMMENT"]

            _dat_cov.update(
                dict(
                    COMMENT=v["COMMENT"],
                    EPOCH=Time(v["EPOCH"], format="datetime64", scale="utc").CCSDS_epoch,
                    COV_REF_FRAME=_meta["REF_FRAME"],
                    CX_X=v["CX_X"],
                    CY_X=v["CY_X"],
                    CY_Y=v["CY_Y"],
                    CZ_X=v["CZ_X"],
                    CZ_Y=v["CZ_Y"],
                    CZ_Z=v["CZ_Z"],
                    CX_DOT_X=v["CX_DOT_X"],
                    CX_DOT_Y=v["CX_DOT_Y"],
                    CX_DOT_Z=v["CX_DOT_Z"],
                    CX_DOT_X_DOT=v["CX_DOT_X_DOT"],
                    CY_DOT_X=v["CY_DOT_X"],
                    CY_DOT_Y=v["CY_DOT_Y"],
                    CY_DOT_Z=v["CY_DOT_Z"],
                    CY_DOT_X_DOT=v["CY_DOT_X_DOT"],
                    CY_DOT_Y_DOT=v["CY_DOT_Y_DOT"],
                    CZ_DOT_X=v["CZ_DOT_X"],
                    CZ_DOT_Y=v["CZ_DOT_Y"],
                    CZ_DOT_Z=v["CZ_DOT_Z"],
                    CZ_DOT_X_DOT=v["CZ_DOT_X_DOT"],
                    CZ_DOT_Y_DOT=v["CZ_DOT_Y_DOT"],
                    CZ_DOT_Z_DOT=v["CZ_DOT_Z_DOT"],
                )
            )

            _data["covarianceMatrix"].append(_dat_cov)

    # serialize to xml
    schema = get_schema()
    xml_etree = schema.encode(d, path="/oem")

    if file is not None:
        file.write(xmlschema.etree_tostring(xml_etree, encoding="unicode", method="xml"))

    return xml_etree


###############################################################################
# TDM
###############################################################################

_TDM_METADATA_FIELDS = [
    "COMMENT",
    "TRACK_ID",
    "TRACK_ID",
    "DATA_TYPES",
    "TIME_SYSTEM",
    "START_TIME",
    "STOP_TIME",
    "PARTICIPANT_1",
    "PARTICIPANT_2",
    "PARTICIPANT_3",
    "PARTICIPANT_4",
    "PARTICIPANT_5",
    "MODE",
    "PATH",
    "EMPHEMERIS_NAME_1",
    "EMPHEMERIS_NAME_2",
    "EMPHEMERIS_NAME_3",
    "EMPHEMERIS_NAME_4",
    "EMPHEMERIS_NAME_5",
    "TRANSMIT_BAND",
    "RECEIVE_BAND",
    "TURNAROUND_NUMENATOR",
    "TURNAROUND_DENUMENATOR",
    "TIMETAG_REF",
    "INTEGRATION_INTERVAL",
    "INTEGRATION_REF",
    "FREQ_OFFSET",
    "RANGE_MODE",
    "RANGE_MODULUS",
    "RANGE_UNITS",
    "ANGLE_TYPE",
    "REFERENCE_FRAME",
    "INTERPOLATION",
    "INTERPOLATION_DEGREE",
    "DOPPLER_COUNT_BIAS",
    "DOPPLER_COUNT_SCALE",
    "DOPPLER_COUNT_ROLLOVER",
    "TRANSMIT_DELAY_1",
    "TRANSMIT_DELAY_2",
    "TRANSMIT_DELAY_3",
    "TRANSMIT_DELAY_4",
    "TRANSMIT_DELAY_5",
    "RECEIVE_DELAY_1",
    "RECEIVE_DELAY_2",
    "RECEIVE_DELAY_3",
    "RECEIVE_DELAY_4",
    "RECEIVE_DELAY_5",
    "DATA_QUALITY",
    "CORRECTION_ANGLE_1",
    "CORRECTION_ANGLE_2",
    "CORRECTION_DOPPLER",
    "CORRECTION_MAG",
    "CORRECTION_RANGE",
    "CORRECTION_RCS",
    "CORRECTION_RECEIVE",
    "CORRECTION_TRANSMIT",
    "CORRECTION_ABERRATION_YEARLY",
    "CORRECTION_ABERRATION_DIURNAL",
    "CORRECTIONS_APPLIED",
]

_TDM_DATA_FIELDS = [
    "ANGLE_1",
    "ANGLE_2",
    "CARRIER_POWER",
    "CLOCK_BIAS",
    "CLOCK_DRIFT",
    "DOPPLER_COUNT",
    "DOPPLER_INSTANTANEOUS",
    "DOPPLER_INTEGRATED",
    "DOR",
    "MAG",
    "PC_N0",
    "PR_N0",
    "PRESSURE",
    "RANGE",
    "RCS",
    "RECEIVE_FREQ",
    "RECEIVE_FREQ_1",
    "RECEIVE_FREQ_2",
    "RECEIVE_FREQ_3",
    "RECEIVE_FREQ_4",
    "RECEIVE_FREQ_5",
    "RECEIVE_PHASE_CT_1",
    "RECEIVE_PHASE_CT_2",
    "RECEIVE_PHASE_CT_3",
    "RECEIVE_PHASE_CT_4",
    "RECEIVE_PHASE_CT_5",
    "RHUMIDITY",
    "STEC",
    "TEMPERATURE",
    "TRANSMIT_FREQ_1",
    "TRANSMIT_FREQ_2",
    "TRANSMIT_FREQ_3",
    "TRANSMIT_FREQ_4",
    "TRANSMIT_FREQ_5",
    "TRANSMIT_FREQ_RATE_1",
    "TRANSMIT_FREQ_RATE_2",
    "TRANSMIT_FREQ_RATE_3",
    "TRANSMIT_FREQ_RATE_4",
    "TRANSMIT_FREQ_RATE_5",
    "TRANSMIT_PHASE_CT_1",
    "TRANSMIT_PHASE_CT_2",
    "TRANSMIT_PHASE_CT_3",
    "TRANSMIT_PHASE_CT_4",
    "TRANSMIT_PHASE_CT_5",
    "TROPO_DRY",
    "TROPO_WET",
    "VLBI_DELAY",
]


###############################################################################
# TDM READ
###############################################################################


def read_xml_tdm(xml):
    """
    read xml TMD file resource, parse and validate according to schema,
    returns python dictionary representing document
    """
    schema = get_schema()
    d = schema.decode(xml)
    return d


###############################################################################
# TDM WRITE
###############################################################################


def write_xml_tdm(data, meta, file=None):
    """
    creaate and validate xml for TDM message
    write xml file resource

    assumes tai input time.
    """
    # originator
    originator = meta.get("ORIGINATOR", f"SORTS {__version__}")

    # creation data
    creation_date = meta.get("CREATION_DATE", Time.now()).CCSDS_epoch

    # message id
    message_id = meta.get("MESSAGE_ID", 0)

    # tdm dictionary
    d = {
        "@id": "CCSDS_TDM_VERS",
        "@version": "2.0",
        "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "@xsi:noNamespaceSchemaLocation": _SCHEMA_URI,
        "header": {
            "CREATION_DATE": f"{creation_date}",
            "ORIGINATOR": f"{originator}",
            "MESSAGE_ID": f"{message_id}",
        },
        "body": {"segment": []},
    }

    # ALTERNATIVES
    #
    # 1) - make one segment per observation or
    # 2) - one segment with multiple observations
    #
    # CHOOSE 2)
    segment = {"metadata": {}, "data": {}}
    d["body"]["segment"].append(segment)

    ###########################################################################
    # SEGMENT METADATA
    ###########################################################################

    _meta = segment["metadata"]

    # mandatory
    meta_defaults = {"TIME_SYSTEM": "UTC", "PARTICIPANT_1": "missing"}

    # add values from given meta - in correct order
    for field in _TDM_METADATA_FIELDS:
        if field in meta:
            _meta[field] = meta[field]
        else:
            if field in meta_defaults:
                _meta[field] = meta_defaults[field]

    ###########################################################################
    # SEGMENT DATA
    ###########################################################################

    _data = segment["data"]
    if "DATA_COMMENT" in meta:
        _data["COMMENT"] = meta["DATA_COMMENT"]

    _data["observation"] = []

    for entry in data:
        for key in _TDM_DATA_FIELDS:
            if key in data.dtype.names:
                _dat = {
                    "EPOCH": Time(entry["EPOCH"], scale="tai", format="datetime64").CCSDS_epoch,
                }
                _dat[key] = entry[key]
                _data["observation"].append(_dat)

    # serialize to xml
    schema = get_schema()
    xml_etree = schema.encode(d, path="/tdm")

    if file is not None:
        file.write(xmlschema.etree_tostring(xml_etree, encoding="unicode", method="xml"))

    return xml_etree


def read_txt_oem(fname):
    """Read the txt format of oem, TODO: generalize and document"""
    meta = {"COMMENT": ""}

    _dtype = [
        ("date", "datetime64[us]"),
        ("x", "float64"),
        ("y", "float64"),
        ("z", "float64"),
        ("vx", "float64"),
        ("vy", "float64"),
        ("vz", "float64"),
    ]

    raw_data = []
    DATA_ON = False
    META_ON = False
    with open(fname, "r") as f:
        for line in f:
            if META_ON:
                if line.strip() == "META_STOP":
                    META_ON = False
                    DATA_ON = True
                    continue

                _tmp = [x.strip() for x in line.split("=")]
                meta[_tmp[0]] = _tmp[1]
            elif DATA_ON:
                if line[:7] == "COMMENT":
                    meta["COMMENT"] += line[7:]
                else:
                    raw_data.append(line.split(" "))
            else:
                if line.strip() == "META_START":
                    META_ON = True
                    continue
                _tmp = [x.strip() for x in line.split("=")]
                meta[_tmp[0]] = _tmp[1]

    data_len = len(raw_data)

    data = np.empty((data_len,), dtype=_dtype)

    for ind, row in enumerate(raw_data):
        rown = 0
        for col, dtype in _dtype:
            data[ind][col] = row[rown]
            rown += 1

    return data, meta


def read_txt_tdm(fname):
    """Read the txt format of tdm, TODO: generalize and document"""
    meta = {"COMMENT": ""}

    RANGE_UNITS = "km"
    with open(fname, "r") as f:
        DATA_ON = False
        META_ON = False
        data_raw = {}
        for line in f:
            if line.strip() == "DATA_STOP":
                break

            if META_ON:
                tmp_lin = line.split("=")
                if len(tmp_lin) > 1:
                    meta[tmp_lin[0].strip()] = tmp_lin[1].strip()
                    if tmp_lin[0].strip() == "RANGE_UNITS":
                        RANGE_UNITS = tmp_lin[1].strip().lower()
            elif DATA_ON:
                name, tmp_dat = line.split("=")

                name = name.strip().lower()
                tmp_dat = tmp_dat.strip().split(" ")

                if name in data_raw:
                    data_raw[name].append(tmp_dat)
                else:
                    data_raw[name] = [tmp_dat]
            else:
                if line.lstrip()[:7] == "COMMENT":
                    meta["COMMENT"] += line.lstrip()[7:]
                else:
                    tmp_lin = line.split("=")
                    if len(tmp_lin) > 1:
                        meta[tmp_lin[0].strip()] = tmp_lin[1].strip()

            if line.strip() == "META_START":
                META_ON = True
            if line.strip() == "DATA_START":
                META_ON = False
                DATA_ON = True
    _dtype = [
        ("date", "datetime64[us]"),
    ]

    data_len = len(data_raw[list(data_raw.keys())[0]])

    for name in data_raw:
        _dtype.append((name, "float64"))
        _dtype.append((name + "_err", "float64"))

    data = np.empty((data_len,), dtype=_dtype)

    date_set = False
    for name, series in data_raw.items():
        for ind, val in enumerate(series):
            if not date_set:
                data[ind]["date"] = np.datetime64(val[0], "us")

            data[ind][name] = np.float64(val[1])
            if len(val) > 2:
                data[ind][name + "_err"] = np.float64(val[2])
            else:
                data[ind][name + "_err"] = 0.0

            if name == "range":
                if RANGE_UNITS == "s":
                    data[ind][name] *= consts.c * 1e-3

        date_set = True

    return data, meta


def write_txt_tdm(data, meta, file=None):
    """
    creaate text string for TDM message

    assumes tai input time.
    """
    # xml_tdm = write_xml_tdm(data, meta, file=None)
    raise NotImplementedError("Rip off the bandied and fix this -__-")

    # with open(file, "w") as fo:
    #     fo.write("CCSDS_TDM_VERS = 1.0\n")
    #     fo.write("   COMMENT MASTER ID %s\n" % (oid))
    #     fo.write("   COMMENT TX_ECEF (%1.12f,%1.12f,%1.12f)\n" % (tx_ecef[0], tx_ecef[1], tx_ecef[2]))
    #     fo.write("   COMMENT RX_ECEF (%1.12f,%1.12f,%1.12f)\n" % (rx_ecef[0], rx_ecef[1], rx_ecef[2]))
    #     fo.write("   COMMENT This is a simulated %s of MASTER ID %s with EISCAT 3D\n" % (tdm_type, oid))
    #     fo.write("   COMMENT 233 MHz, time of flight, with ionospheric corrections\n")
    #     fo.write("   COMMENT EISCAT 3D coordinates: \n")
    #     fo.write("   COMMENT Author(s): SORTS.\n")
    #     fo.write("   CREATION_DATE       = %s\n" % (dates.unix_to_datestr(time.time())))
    #     fo.write("META_START\n")
    #     fo.write("   TIME_SYSTEM         = UTC\n")
    #     fo.write("   START_TIME          = %s\n" % (dates.unix_to_datestrf(t_pulse.min())))
    #     fo.write("   STOP_TIME           = %s\n" % (dates.unix_to_datestrf(t_pulse.max())))
    #     fo.write("   PARTICIPANT_1       = %s\n" % (tx_name))
    #     fo.write("   PARTICIPANT_2       = %s\n" % (oid))
    #     fo.write("   PARTICIPANT_3       = %s\n" % (rx_name))
    #     fo.write("   MODE                = SEQUENTIAL\n")
    #     fo.write("   PATH                = 1,2,3\n")
    #     fo.write("   TRANSMIT_BAND       = %1.5f\n" % (freq))
    #     fo.write("   RECEIVE_BAND        = %1.5f\n" % (freq))
    #     fo.write("   TIMETAG_REF         = TRANSMIT\n")
    #     fo.write("   INTEGRATION_REF     = START\n")
    #     fo.write("   RANGE_MODE          = CONSTANT\n")
    #     fo.write("   RANGE_MODULUS       = %1.2f\n" % (128.0 * 20e-3))
    #     fo.write("   RANGE_UNITS         = KM\n")
    #     fo.write("   DATA_QUALITY        = VALIDATED\n")
    #     fo.write("   CORRECTION_RANGE    = 0.0\n")
    #     fo.write("   NOISE_DATA          = ON\n")
    #     fo.write("   CORRECTIONS_APPLIED = NO\n")
    #     fo.write("META_STOP\n")
    #     fo.write("DATA_START\n")
    #     for ri in range(len(t_pulse)):
    #         fo.write(
    #             "   RANGE                   = %s %1.12f %1.12f\n"
    #             % (dates.unix_to_datestrf(t_pulse[ri]), m_range[ri], m_range_std[ri])
    #         )
    #         fo.write(
    #             "   DOPPLER_INSTANTANEOUS   = %s %1.12f %1.12f\n"
    #             % (dates.unix_to_datestrf(t_pulse[ri]), m_range_rate[ri], m_range_rate_std[ri])
    #         )
    #     fo.write("DATA_STOP\n")



#!/usr/bin/env python

"""Provides functions for read and writing data in standardized formats such as the CCSDS standards.

"""

from .ccsds import write_xml_tdm, read_xml_tdm
from .ccsds import write_xml_oem, read_xml_oem
from .ccsds import read_txt_oem, read_txt_tdm
from .ccsds import write_txt_tdm
from .terminal import flush_table, step_flush_table



#!/usr/bin/env python

"""Convenience functions for interacting with the Linux terminal

"""
import time
import sys

from tabulate import tabulate

ERASE_LINE = "\x1b[2K"
CURSOR_UP_ONE = "\x1b[1A"


def flush_table(table_data, header, chunksize=10, step_size=10, step_time=0.2):
    """ "Animates" a table by flushing multiple chunks of that table to the terminal with a sleep time between flushes."""

    up_flush = CURSOR_UP_ONE + ERASE_LINE

    lines = chunksize + 3
    for i in range(0, len(table_data), step_size):
        if step_time is not None:
            time.sleep(step_time)
        sub_table = table_data[i : (i + chunksize)]

        str_data = "\n" + tabulate(sub_table, headers=header) + "\n"
        if i > 0:
            for j in range(lines):
                sys.stdout.write(up_flush)
                sys.stdout.flush()

        sys.stdout.write(str_data)

        # last possibly un-even table
        if len(sub_table) < chunksize:
            sys.stdout.write("\n " * (chunksize - len(sub_table)))

        sys.stdout.flush()


def step_flush_table(table_data, header, table_size, step_time=None, first_step=False):
    """ "Animates" a table by flushing multiple chunks of that table to the terminal with a sleep time between flushes."""

    up_flush = CURSOR_UP_ONE + ERASE_LINE

    lines = table_size + 3
    if step_time is not None:
        time.sleep(step_time)

    # last possibly un-even table
    if len(table_data) < table_size:
        fill = "\n" * (table_size - len(table_data))
    else:
        fill = ""

    str_data = "\n" + tabulate(table_data, headers=header) + fill + "\n"

    if not first_step:
        for j in range(lines):
            sys.stdout.write(up_flush)
            sys.stdout.flush()

    sys.stdout.write(str_data)

    sys.stdout.flush()



#!/usr/bin/env python

"""

"""

import numpy as np

from .scan import Scan


class Beampark(Scan):
    """Static beampark."""

    def __init__(self, azimuth=0.0, elevation=90.0, dwell=0.1):
        super().__init__(coordinates="azelr")
        self.elevation = elevation
        self.azimuth = azimuth
        self._dwell = dwell

    def dwell(self, t=None):
        if t is None:
            return self._dwell
        else:
            if isinstance(t, float) or isinstance(t, int):
                return self._dwell
            else:
                return np.ones(t.shape, dtype=t.dtype) * self._dwell

    def min_dwell(self):
        return self._dwell

    def cycle(self):
        return None

    def pointing(self, t):
        if isinstance(t, float) or isinstance(t, int):
            shape = (3,)
        else:
            shape = (3, len(t))

        el_len = True
        try:
            shape += (len(self.elevation),)
        except TypeError:
            el_len = False

        az_len = True
        try:
            shape += (len(self.azimuth),)
        except TypeError:
            az_len = False

        azelr = np.empty(shape, dtype=np.float64)

        if az_len:
            for ind in range(len(self.azimuth)):
                if len(shape) == 2:
                    azelr[0, ind] = self.azimuth[ind]
                else:
                    azelr[0, :, ind] = self.azimuth[ind]
        else:
            azelr[0, ...] = self.azimuth

        if el_len:
            for ind in range(len(self.elevation)):
                if len(shape) == 2:
                    azelr[1, ind] = self.elevation[ind]
                else:
                    azelr[1, :, ind] = self.elevation[ind]
        else:
            azelr[1, ...] = self.elevation

        azelr[2, ...] = 1.0
        return azelr



from . import simulation_unit, stx_mrx_simulation
from .stx_mrx_simulation import (
    SpaceObjectDsecSampler,
    sample_and_propagate_space_objects_states,
    group_passages_by_tx_rx_station_pair,
    find_passages,
    iter_mpi_simulation_results,
    StxMrxSimulation,
)
from .simulation_unit import (
    SimulationUnit,
    Observation,
    SimulationUnitState,
    FromPassagesOverTxRxStationPairParam,
)



#!/usr/bin/env python

"""#TODO

"""

import logging
import numpy as np

from .radar_controller import RadarController

logger = logging.getLogger(__name__)

class Scanner(RadarController):
    """Takes in a scan and create a scanning radar controller."""

    META_FIELDS = RadarController.META_FIELDS + [
        "scan_type",
        "dwell",
    ]

    def __init__(
        self,
        radar,
        scan,
        t0=0.0,
        r=np.linspace(300e3, 1000e3, num=10),
        as_altitude=False,
        return_copy=False,
        meta=None,
        **kwargs
    ):
        super().__init__(radar, t0=t0, meta=meta, **kwargs)
        self.scan = scan
        if self.t is not None and self.t_slice is None:
            self.dwell = np.max(self.scan.dwell(self.t))

        self.r = r
        self.return_copy = return_copy
        self.as_altitude = as_altitude

        logger.info(f"Scanner:init")

    @property
    def dwell(self):
        if self.t_slice is None:
            return self.scan.dwell(self.t)
        else:
            return self.t_slice

    @dwell.setter
    def dwell(self, val):
        self.t_slice = val

    def default_meta(self):
        dic = super().default_meta()
        dic["scan_type"] = self.scan.__class__
        return dic

    def point_radar(self, t):
        """Assumes t is not array"""

        if self.return_copy:
            radar = self.radar.copy()
        else:
            radar = self.radar

        meta = self.default_meta()
        meta["dwell"] = self.scan.dwell(t)

        RadarController.coh_integration(radar, meta["dwell"])

        point_rx_to_tx = []
        point_tx = []
        for tx in radar.tx:
            point = self.scan.ecef_pointing(t, tx)

            if self.as_altitude:
                if len(point.shape) > 1:
                    r = self.r[None, :] / point[2, :]
                    point_tx.append(point + tx.ecef[:, None])
                    __ptx = point[:, :, None] * r[None, :, :] + tx.ecef[:, None, None]
                    point_rx_to_tx.append(__ptx.reshape(3, __ptx.shape[1] * __ptx.shape[2]))
                else:
                    r = self.r / point[2]
                    point_tx.append(point + tx.ecef)
                    point_rx_to_tx.append(point[:, None] * r[None, :] + tx.ecef[:, None])
            else:
                if len(point.shape) > 1:
                    point_tx.append(point + tx.ecef[:, None])
                    __ptx = point[:, :, None] * self.r[None, None, :] + tx.ecef[:, None, None]
                    point_rx_to_tx.append(__ptx.reshape(3, __ptx.shape[1] * __ptx.shape[2]))
                else:
                    point_tx.append(point + tx.ecef)
                    point_rx_to_tx.append(point[:, None] * self.r[None, :] + tx.ecef[:, None])

            RadarController._point_station(tx, point_tx[-1])

        for rx in radar.rx:
            rx_point = []
            for txi, tx in enumerate(radar.tx):
                # < 200 meters apart = same location for pointing
                if np.linalg.norm(tx.ecef - rx.ecef) < 200.0:
                    __ptx = point_tx[txi]
                    if len(__ptx.shape) == 1:
                        __ptx = __ptx.reshape(3, 1)
                    rx_point.append(__ptx)
                else:
                    rx_point.append(point_rx_to_tx[txi])
            rx_point = np.concatenate(rx_point, axis=1)

            if len(rx_point.shape) > 1 and rx_point.size == 3:
                rx_point.shape = (3,)

            RadarController._point_station(rx, rx_point)

        # Make sure radar is on
        self.toggle_stations(t, radar)

        return radar, meta

    def generator(self, t):
        for ti in range(len(t)):
            yield self.point_radar(t[ti])



#!/usr/bin/env python

"""

"""

import numpy as np

from .scan import Scan


class Uniform(Scan):
    """Uniformly sample the FOV using Fibonacci lattice."""

    def __init__(self, min_elevation=30.0, dwell=0.1, sph_num=1000):
        super().__init__(coordinates="enu")
        self._dwell = dwell
        self.min_elevation = min_elevation
        self.num = sph_num
        self.sph_num = sph_num

        k = np.empty((3, self.num), dtype=np.float64)

        golden_ratio = (1 + 5**0.5) / 2
        inds = np.arange(0, self.num)
        theta = 2 * np.pi * inds / golden_ratio
        phi = np.arccos(1 - 2 * (inds + 0.5) / self.num)

        k[0, :] = np.cos(theta) * np.sin(phi)
        k[1, :] = np.sin(theta) * np.sin(phi)
        k[2, :] = np.cos(phi)

        min_z = np.sin(np.radians(min_elevation))
        k = k[:, k[2, :] >= min_z]
        self.num = k.shape[1]
        self.pointings = k

    def dwell(self, t=None):
        if t is None:
            return self._dwell
        else:
            if isinstance(t, float) or isinstance(t, int):
                return self._dwell
            else:
                return np.ones(t.shape, dtype=t.dtype) * self._dwell

    def min_dwell(self):
        return self._dwell

    def cycle(self):
        return self.num * self._dwell

    def pointing(self, t):
        ind = (np.mod(t / self.cycle(), 1) * self.num).astype(np.int64)
        return self.pointings[:, ind]



#!/usr/bin/env python

"""Radar configuration plot functions

"""

# Python standard import

# Third party import
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D

# Local import

from . import general
from .. import frames


def radar_earth(ax, radar, **kwargs):
    tx_names = kwargs.pop("tx_names", None)
    rx_names = kwargs.pop("rx_names", None)
    grid_earth(ax, **kwargs)

    for ind, tx in enumerate(radar.tx):
        if tx_names is None:
            ax.plot([tx.ecef[0]], [tx.ecef[1]], [tx.ecef[2]], "x", color="r")
        else:
            ax.plot([tx.ecef[0]], [tx.ecef[1]], [tx.ecef[2]], "x", color="r", label=tx_names[ind])

    for ind, rx in enumerate(radar.rx):
        if rx_names is None:
            ax.plot([rx.ecef[0]], [rx.ecef[1]], [rx.ecef[2]], "x", color="r")
        else:
            ax.plot([rx.ecef[0]], [rx.ecef[1]], [rx.ecef[2]], "x", color="r", label=rx_names[ind])


def radar_map(radar, ax=None):
    """Plot the geographical location of the radar system using the GeoPandas library.

    To get:

    pip install git+git://github.com/geopandas/geopandas.git
    pip install descartes

    """

    try:
        import geopandas
    except ImportError:
        geopandas = None
        raise ImportError('Cannot plot geo-location of radar without "geopandas" package')

    if ax is None:
        fig = plt.figure(figsize=(15, 7))
        ax = fig.add_subplot(111)

    df = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    ax = df.plot(
        figsize=(10, 10), alpha=0.5, edgecolor="k", linewidth=0.25, color="lightgrey", ax=ax
    )

    max_min_geo = {"lat": [0, 0], "lon": [0, 0]}
    sites = []
    lats = []
    lons = []
    for ind, r in enumerate(radar.tx):
        lons.append(r.lon)
        lats.append(r.lat)
        sites.append(f"TX-{ind}")
    df_sites_tx = pd.DataFrame({"Site": sites, "Latitude": lats, "Longitude": lons})
    sites_rx = []
    lats_rx = []
    lons_rx = []
    for ind, r in enumerate(radar.rx):
        lons_rx.append(r.lon)
        lats_rx.append(r.lat)
        sites_rx.append(f"RX-{ind}")
    df_sites_rx = pd.DataFrame({"Site": sites_rx, "Latitude": lats_rx, "Longitude": lons_rx})

    max_min_geo["lat"][0] = np.min(np.array([lats + lats_rx]))
    max_min_geo["lat"][1] = np.max(np.array([lats + lats_rx]))
    max_min_geo["lon"][0] = np.min(np.array([lons + lons_rx]))
    max_min_geo["lon"][1] = np.max(np.array([lons + lons_rx]))
    _lat_df = 5.0
    _lon_df = 10.0

    if max_min_geo["lat"][0] > -(90.0 - _lat_df):
        max_min_geo["lat"][0] -= _lat_df
    if max_min_geo["lat"][1] < 90.0 - _lat_df:
        max_min_geo["lat"][1] += _lat_df

    if max_min_geo["lon"][0] > -(180.0 - _lon_df):
        max_min_geo["lon"][0] -= _lon_df
    if max_min_geo["lon"][1] < 180.0 - _lon_df:
        max_min_geo["lon"][1] += _lon_df

    _font = 8

    df_sites_rx["Coordinates"] = list(zip(df_sites_rx.Longitude, df_sites_rx.Latitude))
    df_sites_rx["Coordinates"] = df_sites_rx["Coordinates"].apply(Point)
    gdf_rx = geopandas.GeoDataFrame(df_sites_rx, geometry="Coordinates")
    gdf_rx.plot(ax=ax, color="blue", alpha=0.3, marker="h", markersize=24)
    for ind in range(gdf_rx.shape[0]):
        plt.text(
            gdf_rx["Coordinates"][ind].x + 0.1,
            gdf_rx["Coordinates"][ind].y + 0.2,
            sites_rx[ind],
            fontsize=_font,
        )

    df_sites_tx["Coordinates"] = list(zip(df_sites_tx.Longitude, df_sites_tx.Latitude))
    df_sites_tx["Coordinates"] = df_sites_tx["Coordinates"].apply(Point)
    gdf = geopandas.GeoDataFrame(df_sites_tx, geometry="Coordinates")
    gdf.plot(ax=ax, color="red", alpha=0.7, marker="X", markersize=18)
    for ind in range(gdf.shape[0]):
        plt.text(
            gdf["Coordinates"][ind].x + 0.1,
            gdf["Coordinates"][ind].y + 0.5,
            sites[ind],
            fontsize=_font,
        )

    ax.set_title("Geographical location for: {}".format(radar.name))
    ax.set_xlim(*max_min_geo["lon"])
    ax.set_ylim(*max_min_geo["lat"])

    ax.set_aspect("equal")

    return ax



from __future__ import annotations
import typing as t
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from sorts import types, radar, scheduling
from sorts.types import TxRxTuple
from sorts.utils import assert_class_attributes_equal_to, to_datetime64_us
from sorts.space_object import SpaceObject
from sorts.radar import Station
from sorts.signals import hard_target_snr
from sorts.interpolation import Interpolator
from sorts.scheduling import ExperimentDetailMap, Schedule
from sorts.simulation import Passage


CoordKey = t.Literal["multi_index", "time", "exp_num", "rx_simult_num", "enu", "e", "n", "u"]
DataKey = t.Literal[
    "tx_pointing",
    "rx_pointing",
    "exp_num",
    "gain_tx",
    "gain_rx",
    "snr",
    "tx_range",
    "rx_range",
    "two_way_range",
    "two_way_range_rate",
]
Key = t.Literal[DataKey, CoordKey]


# TODO: updated the name with tx/rx as suffix to prefix
class _K:
    """Internal helper for accessing string keys consistently"""

    multi_index: t.Final = "multi_index"
    time: t.Final = "time"
    exp_num: t.Final = "exp_num"
    rx_simult_num: t.Final = "rx_simult_num"
    enu: t.Final = "enu"
    e: t.Final = "e"
    n: t.Final = "n"
    u: t.Final = "u"
    tx_pointing: t.Final = "tx_pointing"
    rx_pointing: t.Final = "rx_pointing"
    gain_tx: t.Final = "gain_tx"
    gain_rx: t.Final = "gain_rx"
    snr: t.Final = "snr"
    tx_range: t.Final = "tx_range"
    rx_range: t.Final = "rx_range"
    two_way_range: t.Final = "two_way_range"
    two_way_range_rate: t.Final = "two_way_range_rate"


assert_class_attributes_equal_to(_K, t.get_args(Key))

_SK = scheduling._K
"""Internal helper for accessing string keys consistently"""

SimulationUnitState = t.NewType("SimulationUnitState", xr.Dataset)
"""
A xarray `Dataset` with:
  ```
  Dimensions:        (multi_index: n, enu: 3)
  Coordinates:
    * multi_index    (multi_index) object MultiIndex ('exp_num', 'rx_simult_num', 'time')
    * time           (multi_index) datetime64[us]
    * exp_num        (multi_index) int16
    * rx_simult_num  (multi_index) int16
    * enu            (enu) 'e' 'n' 'u'
  Data variables:
      tx_pointing    (enu, multi_index) float64
      rx_pointing    (enu, multi_index) float64
      gain_tx        (multi_index) float64
      gain_rx        (multi_index) float64
      snr            (multi_index) float64
      tx_range       (multi_index) float64
      rx_range       (multi_index) float64
      two_way_range  (multi_index) float64
      two_way_range_rate  (multi_index) float64
  ```
"""


def empty_state() -> SimulationUnitState:
    multi_index = pd.MultiIndex.from_arrays(
        [
            np.empty(0, dtype=np.int16),
            np.empty(0, dtype=np.int16),
            np.empty(0, dtype="datetime64[us]"),
        ],
        names=(_K.exp_num, _K.rx_simult_num, _K.time),
    )

    state = xr.Dataset(
        coords={
            **xr.Coordinates.from_pandas_multiindex(multi_index, _K.multi_index),
            _K.enu: [_K.e, _K.n, _K.u],
        },
        data_vars={
            # NOTE: we used `.loc` instead of `reindex` here because we cannot get `reindex` working
            # TODO: investigate why `reindex` won't work
            #   not working: `tx_sch[_SK.pointing].reindex({_SK.multi_index: [(np.datetime64("2025-01-01 02:45:01", "us"), 0, 0), ...]})`
            _K.tx_pointing: (
                (_K.enu, _K.multi_index),
                np.empty((3, 0), dtype=np.float64),
            ),
            _K.rx_pointing: ((_K.enu, _K.multi_index), np.empty((3, 0), dtype=np.float64)),
        },
    )

    return SimulationUnitState(state)


def filter_state_by_time_range(
    state: SimulationUnitState, time_range: types.TimeRange_us
) -> SimulationUnitState:
    mask = (state[_K.time] >= time_range[0]) & (state[_K.time] <= time_range[1])
    state_masked = state[{_K.multi_index: mask}]

    return state_masked


def calc_gain(
    state: SimulationUnitState,
    tx_stn: Station,
    rx_stn: Station,
    spobj_tx_enu: types.EnuCoordinates,
    spobj_rx_enu: types.EnuCoordinates,
) -> SimulationUnitState:
    # will be populated to [tx_gain_arr, rx_gain_arr]
    gain_arr_list: list[npt.NDArray[np.float64]] = []

    for stn, spobj_stn_enu, pt_key in zip(
        [tx_stn, rx_stn],
        [spobj_tx_enu, spobj_rx_enu],
        [_K.tx_pointing, _K.rx_pointing],
    ):
        # early return for empty cases
        # NOTE: this is particularly needed because some `.gain` does not work with empty parameters (e.g. beam.parameters["pointing"])
        # TODO: add test case for empty case?
        if len(state[_K.multi_index]) == 0:
            gain_arr_list.append(np.empty(0, dtype=np.float64))

        elif stn.beam_parameters is None:
            # TODO: remove this hack; see issues #25 for details
            raise RuntimeError(
                "A hack of injecting `beam_parameters` into `tx_stn.beam_parameters` is currently required for gain calculation"
            )

        else:
            beam_parameters = stn.beam_parameters

            if "pointing" in stn.beam_parameters.keys:
                beam_parameters = stn.beam_parameters.replace_and_broadcast(
                    parameters=stn.beam_parameters,
                    new_parameters=dict(pointing=state[pt_key].to_numpy()),
                )

            gain_arr_list.append(stn.beam.gain(spobj_stn_enu[:3], beam_parameters))

    state[_K.gain_tx] = (_K.multi_index, gain_arr_list[0])
    state[_K.gain_rx] = (_K.multi_index, gain_arr_list[1])

    return state


@dataclass
class FromPassagesOverTxRxStationPairParam:
    id: str
    passages: list[Passage]
    spobj: SpaceObject
    spobj_interp: Interpolator
    tx_station: Station
    rx_station: Station
    schedule: Schedule
    exp_detail_map: ExperimentDetailMap


# TODO: re-eval: `Station`` can be taken from `Passage`, but empty `list[Passage]` would be an issue in that case.
class SimulationUnit:
    """
    Contains all the params and results for a unit of simulation calculation.

    Notes about the state data:
    - It is stored in a private attribute `_state`
    - It can contain data for more than 1 passage
    - The dataset does not always contains all the key defined in `DataKey`,
      which ones are available depends on what calculation have been done.
    """

    _K = _K
    """shortcut to module attribute"""

    # todo: my type checker is complaining over these shortcuts?
    # FromPassagesOverTxRxStationPairParam = FromPassagesOverTxRxStationPairParam
    """shortcut to module attribute"""

    def __init__(
        self,
        id: str,
        spobj: SpaceObject,
        spobj_interp: Interpolator,
        passages: list[Passage],
        tx_station: Station,
        rx_station: Station,
        exp_detail_map: ExperimentDetailMap,
        state: SimulationUnitState,
    ):
        self.id = id

        self._state = state

        self.space_object = spobj
        self.space_object_interp = spobj_interp

        self.passages = passages

        self.tx_station = tx_station
        self.rx_station = rx_station
        self.exp_detail_map = exp_detail_map

        self.observations: list[Observation] = []

    # todo: this param should maybe be expanded so the components are arguments, or a more
    # generalized units should be made: i think this might be too specialized as a data carrier?
    # will it be useful outside of this function call?
    @classmethod
    def from_passages_over_tx_rx_station_pair(
        cls, param: FromPassagesOverTxRxStationPairParam
    ) -> t.Self:
        # TODO: the logic inside this function is not super clear - it needs clarification
        id = param.id
        passages = param.passages
        spobj = param.spobj
        spobj_interp = param.spobj_interp
        tx_station = param.tx_station
        rx_station = param.rx_station

        # early return for empty cases
        # NOTE: this is particularly needed because `.loc` will throw KeyError for non-existence keys
        # TODO: add test case for empty case?
        if (
            len(passages) == 0
            or not (param.schedule[_SK.stn_num] == tx_station.uid).any()
            or not (param.schedule[_SK.stn_num] == rx_station.uid).any()
        ):
            return cls(
                id=id,
                spobj=spobj,
                spobj_interp=spobj_interp,
                passages=passages,
                tx_station=param.tx_station,
                rx_station=param.rx_station,
                exp_detail_map=param.exp_detail_map,
                state=SimulationUnitState(empty_state()),
            )

        # NOTE: xarray simplify/collapse MultiIndex when filtering a level to an exact value,
        #   we filter on the top level "multi_index' with a tuple here to prevent it
        tx_schdata = param.schedule.loc[
            {_SK.multi_index: (slice(None), tx_station.uid, slice(None), slice(None))}
        ]
        rx_schdata = param.schedule.loc[
            {_SK.multi_index: (slice(None), rx_station.uid, slice(None), slice(None))}
        ]

        rx_time = rx_schdata[_SK.start_time].to_numpy()
        rx_exp_num = rx_schdata[_SK.exp_num].to_numpy()
        rx_simult_num = rx_schdata[_SK.simult_num].to_numpy()

        multi_index = pd.MultiIndex.from_arrays(
            [rx_exp_num, rx_simult_num, rx_time],
            names=(_K.exp_num, _K.rx_simult_num, _K.time),
        )

        tx_reindex_selector = xr.Coordinates.from_pandas_multiindex(
            pd.MultiIndex.from_arrays(
                [
                    rx_exp_num,
                    np.full(len(rx_time), param.tx_station.uid, dtype=np.int16),
                    np.full(len(rx_time), 0, dtype=np.int16),  # assuming single tx
                    rx_time,
                ],
                names=(_SK.exp_num, _SK.stn_num, _SK.simult_num, _SK.start_time),
            ),
            _SK.multi_index,
        )

        state = xr.Dataset(
            coords={
                **xr.Coordinates.from_pandas_multiindex(multi_index, _K.multi_index),
                _K.enu: [_K.e, _K.n, _K.u],
            },
            data_vars={
                # NOTE: we used `.loc` instead of `reindex` here because we cannot get `reindex` working
                # TODO: investigate why `reindex` won't work
                #   not working: `tx_sch[_SK.pointing].reindex({_SK.multi_index: [(np.datetime64("2025-01-01 02:45:01", "us"), 0, 0), ...]})`
                _K.tx_pointing: (
                    (_K.enu, _K.multi_index),
                    tx_schdata[_SK.pointing]
                    .loc[{_SK.multi_index: tx_reindex_selector[_SK.multi_index]}]
                    .to_numpy(),
                ),
                _K.rx_pointing: ((_K.enu, _K.multi_index), rx_schdata[_SK.pointing].to_numpy()),
            },
        )

        return cls(
            id=id,
            spobj=spobj,
            spobj_interp=spobj_interp,
            passages=passages,
            tx_station=param.tx_station,
            rx_station=param.rx_station,
            exp_detail_map=param.exp_detail_map,
            state=SimulationUnitState(state),
        )

    def simulate(self):
        """
        Run simulation calculations and update its state/data;
        Will populate the prop `observations`
        """

        if self.tx_station.wavelength is None:
            # TODO: remove this hack; see issues #25 for details
            raise RuntimeError(
                "A hack of injecting `frequency` into `tx_stn.frequency` is currently required for calling `hard_target_snr`"
            )

        epoch = to_datetime64_us(self.space_object.epoch)
        dsec = (self._state[_K.time] - epoch).to_numpy() / np.timedelta64(1, "s")
        spobj_states = self.space_object_interp.get_state(dsec)
        spobj_tx_enu = self.tx_station.enu(spobj_states)
        spobj_rx_enu = self.rx_station.enu(spobj_states)

        range_tx: npt.NDArray[types.Float64_as_m] = np.linalg.norm(spobj_tx_enu[:3, :], axis=0)
        range_rx: npt.NDArray[types.Float64_as_m] = np.linalg.norm(spobj_rx_enu[:3, :], axis=0)

        # TODO: can likely use assignment by slice/indexing instead of looping
        # TODO: do we need `pulse_lengths`?
        # TODO: do we need `ipps`?
        # TODO: do we need `duty_cycles`?
        powers = np.array(
            [self.exp_detail_map[n].power for n in self._state[_K.exp_num].to_numpy()],
            dtype=np.float64,
        )
        bandwidths = np.array(
            [self.exp_detail_map[n].bandwidth for n in self._state[_K.exp_num].to_numpy()],
            dtype=np.float64,
        )
        rx_noise_temps = np.array(
            [self.exp_detail_map[n].noise_temp for n in self._state[_K.exp_num].to_numpy()],
            dtype=np.float64,
        )

        self._state = calc_gain(
            state=self._state,
            tx_stn=self.tx_station,
            rx_stn=self.rx_station,
            spobj_tx_enu=spobj_tx_enu,
            spobj_rx_enu=spobj_rx_enu,
        )

        snr = hard_target_snr(
            gain_tx=self._state[_K.gain_tx].to_numpy(),
            gain_rx=self._state[_K.gain_rx].to_numpy(),
            wavelength=self.tx_station.wavelength,
            power_tx=powers,
            range_tx_m=range_tx,
            range_rx_m=range_rx,
            diameter=self.space_object.d,
            bandwidth=bandwidths,
            rx_noise_temp=rx_noise_temps,
            radar_albedo=self.space_object.properties.get("radar_albedo", 1.0),
        )
        self._state[_K.snr] = (_K.multi_index, snr)

        self._state[_K.tx_range] = (
            _K.multi_index,
            np.linalg.norm(spobj_tx_enu[:3, :], axis=0),
        )

        self._state[_K.rx_range] = (
            _K.multi_index,
            np.linalg.norm(spobj_rx_enu[:3, :], axis=0),
        )

        self._state[_K.two_way_range] = (
            _K.multi_index,
            range_tx + range_rx,
        )
        v_tx = np.sum(spobj_tx_enu[:3, :] * spobj_tx_enu[3:, :], axis=0) / range_tx
        v_rx = np.sum(spobj_rx_enu[:3, :] * spobj_rx_enu[3:, :], axis=0) / range_rx
        self._state[_K.two_way_range_rate] = (
            _K.multi_index,
            v_tx + v_rx,
        )

        obss = self.get_observations()
        self.observations = obss

        return obss

    def get_observations(self) -> list[Observation]:
        obss: list[Observation] = []

        for passage in self.passages:
            obss.extend(Observation.from_passage(passage, self))

        return obss


ObservationStationScheduleIndexer = tuple[
    scheduling.ExperimentId,
    radar.StationId,
    scheduling.SimultaneousNum,
    npt.NDArray[types.Datetime64_us],
]
ObservationScheduleIndexer = TxRxTuple[
    ObservationStationScheduleIndexer, ObservationStationScheduleIndexer
]
ObservationStateIndexer = tuple[
    scheduling.ExperimentId, scheduling.SimultaneousNum, npt.NDArray[types.Datetime64_us]
]


# TODO: i dont understand why this does not seem to actually contain any data? everything seems to
# be in the simulation units? maybe parts of the simulations units could be moved here or vice versa
class Observation:
    def __init__(
        self,
        passage: Passage,
        sim_unit: SimulationUnit,
        exp_id: scheduling.ExperimentId,
        simult_num: scheduling.SimultaneousNum,
    ):
        # todo: update for collecting passage and multi passage
        self.passage = passage
        self.sim_unit = sim_unit
        self.exp_id = exp_id
        self.simult_num = simult_num

    @classmethod
    def from_passage(cls, passage: Passage, sim_unit: SimulationUnit) -> list[t.Self]:
        state_slice = filter_state_by_time_range(sim_unit._state, passage.time_range)

        multi_index = t.cast(pd.MultiIndex, state_slice.indexes[_K.multi_index])

        unique_exp_id_simult_num_pairs: list[
            tuple[scheduling.ExperimentId, scheduling.SimultaneousNum]
        ] = (multi_index.droplevel(_K.time).unique().to_list())

        obss = [
            cls(passage=passage, sim_unit=sim_unit, exp_id=exp_id, simult_num=simult_num)
            for exp_id, simult_num in unique_exp_id_simult_num_pairs
        ]

        return obss

    def __repr__(self) -> str:
        return "\n".join(
            [
                "Observation(",
                f"    time_range={self.passage.time_range}",
                f"    spobj_id={self.sim_unit.space_object.object_id}",
                f"    tx_stn_id={self.passage.tx_station.uid},"
                f"    rx_stn_id={self.passage.rx_stations[0].uid}",
                f"    exp_id={self.exp_id}, simult_num={self.simult_num}",
                ")",
            ]
        )

    def get_time_arr(self):
        time_arr = filter_state_by_time_range(self.sim_unit._state, self.passage.time_range)[
            _K.time
        ].to_numpy()

        return time_arr

    def index_into_schedule(self, sch: Schedule) -> TxRxTuple[Schedule, Schedule]:
        """Returns subset of schedules, in `(tx_scheule, tx_schedule` that corresponds to the observation"""

        tx_sch_obs = scheduling.filter_by_time_range(sch, self.passage.time_range)
        tx_sch_obs = tx_sch_obs.loc[
            {
                _SK.multi_index: (
                    self.exp_id,
                    self.passage.tx_station.uid,
                    0,  # NOTE: we only support single simultaneous tx pointing
                    slice(None),
                )
            }
        ]

        rx_sch_obs = scheduling.filter_by_time_range(sch, self.passage.time_range)
        rx_sch_obs = rx_sch_obs.loc[
            {
                _SK.multi_index: (
                    self.exp_id,
                    self.passage.rx_stations[0].uid,
                    self.simult_num,
                    slice(None),
                )
            }
        ]

        return TxRxTuple(tx=tx_sch_obs, rx=rx_sch_obs)

    def get_state_slice(self) -> SimulationUnitState:
        """Get the subset of `State` data the corresponds to the the observation"""

        sim_state_slice = filter_state_by_time_range(self.sim_unit._state, self.passage.time_range)

        # NOTE: early return for empty case; `loc` method does not work with non-existent selection
        if len(sim_state_slice[_K.multi_index]) == 0:
            return sim_state_slice

        sim_state_slice = sim_state_slice.loc[
            {_K.multi_index: (self.exp_id, self.simult_num, slice(None))}
        ]

        return sim_state_slice



#!/usr/bin/env python

"""Package that defines populations and methods for loading them from different data sources."""

from .population import Population

from .master import master_catalog
from .master import master_catalog_factor

from .tles import tle_catalog

from .minimoon import NESCv9_minimoons

from .grids import orbit_grid



#!/usr/bin/env python

"""Provides a static list of controllers, simplest scheduler.

"""

import logging

from .scheduler import Scheduler

logger = logging.getLogger(__name__)

class StaticList(Scheduler):
    """#TODO: Docstring"""

    def __init__(self, radar, controllers, **kwargs):
        super().__init__(
            radar=radar,
        )
        self.controllers = controllers

    def update(self, controllers):
        logger.debug(f"StaticList:update:id(controllers) = {id(controllers)}")
        self.controllers = controllers

    def get_controllers(self):
        return self.controllers



#!/usr/bin/env python

import numpy as np
import spacecoords
from sorts import frames


def set_station_ecef(station):
    station["ecef"] = frames.geodetic_to_ITRS(
        station["lat"], station["lon"], station["alt"], degrees=True
    )
    ecef_lla = spacecoords.spherical.cart_to_sph(station["ecef"], degrees=True)
    station["ecef_lat"] = ecef_lla[1]
    station["ecef_lon"] = 90 - ecef_lla[0]
    station["ecef_alt"] = ecef_lla[2]


def field_of_view(station, states):
    """Determines the field of view of the station.
    Should be vectorized over second dimension of states.
    Needs to return a numpy boolean array with `True` when the state is inside the FOV.

    Used to determine when a "pass" is occurring based on the input ECEF states and times.
    The default implementation is a minimum elevation check.
    """
    zenith = np.array([0, 0, 1], dtype=np.float64)

    enu_states = enu(station, states[:3, :])

    zenith_ang = spacecoords.linalg.vector_angle(zenith, enu_states, degrees=True)
    check = zenith_ang < 90.0 - station.get("min_elevation", 0)

    return check


def enu(station, ecefs):
    """Converts a set of ECEF states to local ENU coordinates using geocentric zenith."""
    rel_ = ecefs.copy()
    rel_[:3, :] = rel_[:3, :] - station["ecef"][:, None]
    rel_[:3, :] = frames.ecef_to_enu(
        station["ecef_lat"],
        station["ecef_lon"],
        station["ecef_alt"],
        rel_[:3, :],
        degrees=True,
    )
    if ecefs.shape[0] > 3:
        rel_[3:, :] = frames.ecef_to_enu(
            station["ecef_lat"],
            station["ecef_lon"],
            station["ecef_alt"],
            rel_[3:, :],
            degrees=True,
        )
    return rel_


def point_ecef(station, point):
    """Point Station beam in location of ECEF coordinate. Returns local pointing direction."""
    k = frames.ecef_to_enu(
        station["ecef_lat"],
        station["ecef_lon"],
        station["ecef_alt"],
        point,
        degrees=True,
    )
    k_norm = np.linalg.norm(k, axis=0)
    station["beam"].point(k / k_norm)
    return k



#!/usr/bin/env python

"""General plotting functions

"""

# Python standard import
from itertools import combinations

# Third party import
import numpy as np
import matplotlib.pyplot as plt

import pyorb

# Local import
from .. import constants


def kepler_orbit(orb, res=100, ax=None, **kwargs):
    """Plot a 3d kepler orbit from a single pyorb.Orbit item."""
    assert len(orb) == 1, "Can only handle one orbit for now"
    orbc = orb.copy()
    orbc.direct_update = False
    orbc.auto_update = False

    orbc.add(res - 1)
    kep0 = np.reshape(orb.kepler[:, 0], (6, 1))
    orbc.kepler = np.tile(kep0, (1, res))
    mu = np.linspace(0, 2 * np.pi, res)
    if orb.degrees:
        mu = np.degrees(mu)
    orbc.mean_anomaly = mu

    orbc.calculate_cartesian()
    states = orbc.cartesian

    if ax is None:
        fig = plt.figure(figsize=(15, 15))
        ax = fig.add_subplot(111, projection="3d")
    else:
        fig = None
    ax.plot(states[0, :], states[1, :], states[2, :], "-", **kwargs)

    return fig, ax


def kepler_scatter(o, **options):
    """This function creates several scatter plots of a set of orbital elements based on the
    different possible axis planar projections, calculates all possible permutations of plane
    intersections based on the number of columns

    :param numpy.ndarray o: Rows are distinct orbits and columns are orbital elements in the order a, e, i, omega, Omega
    :param options:  dictionary containing all the optional settings

    #TODO: this needs updating

    Currently the options fields are:
        :marker [char]: the marker type
        :size [int]: the size of the marker
        :title [string]: the title of the plot
        :title_font_size [int]: the title font size
        :axis_labels [list of strings]: labels for each column
        :tick_font_size [int]: the axis tick font size
        :window [tuple/list]: the size of the plot window in pixels (assuming dpi = 80)
        :save [string]: will not display figure and will instead save it to this path
        :show [bool]: if False will do draw() instead of show() allowing script to continue
        :tight_rect [list of 4 floats]: configuration for the tight_layout function
        :usetex [bool]: whether to typeset labels using TeX syntax


    Example::

        import dpt_tools as dpt
        import numpy as np
        np.random.seed(19680221)

        orbs = np.matrix([
            11  + 3  *np.random.randn(1000),
            0.5 + 0.2*np.random.randn(1000),
            60  + 10 *np.random.randn(1000),
            120 + 5  *np.random.randn(1000),
            33  + 2  *np.random.randn(1000),
        ]).T

        dpt.orbits(orbs,
            title = "My orbital element distribution",
            size = 10,
        )


    """
    if not isinstance(o, np.ndarray):
        o = np.array(o)

    data_axis = options.get("axis", 0)
    if data_axis == 1:
        dim_axis = 0
    else:
        dim_axis = 1

    scale = options.get("scale", np.ones((o.shape[dim_axis],), dtype=o.dtype))

    # turn on TeX interpreter (or not)
    usetex = options["usetex"] if "usetex" in options else True
    plt.rc("text", usetex=usetex)

    lis = list(range(o.shape[dim_axis]))
    axis_plot = list(combinations(lis, 2))

    axis_labels = options.setdefault("axis_labels", None)

    limits = options.get("limits", None)

    if isinstance(axis_labels, str):
        if axis_labels == "earth-orbit":
            axis_labels = [
                "$a$ [$R_E$]",
                "$e$ [1]",
                "$i$ [deg]",
                "$\omega$ [deg]",
                "$\Omega$ [deg]",
                "$M_0$ [deg]",
            ]
            scale = [1 / constants.R_earth] + [1] * 5
        elif axis_labels == "earth-state":
            axis_labels = [
                "$x$ [$R_E$]",
                "$y$ [$R_E$]",
                "$z$ [$R_E$]",
                "$v_x$ [km/s]",
                "$v_y$ [km/s]",
                "$v_z$ [km/s]",
            ]
            scale = [1 / constants.R_earth] * 3 + [1e-3] * 3
        elif axis_labels == "sol-orbit":
            axis_labels = [
                "$a$ [AU]",
                "$e$ [1]",
                "$i$ [deg]",
                "$\omega$ [deg]",
                "$\Omega$ [deg]",
                "$M_0$ [deg]",
            ]
            scale = [1 / pyorb.AU] + [1] * 5
        elif axis_labels == "sol-state":
            axis_labels = [
                "$x$ [AU]",
                "$y$ [AU]",
                "$z$ [AU]",
                "$v_x$ [km/s]",
                "$v_y$ [km/s]",
                "$v_z$ [km/s]",
            ]
            scale = [1 / pyorb.AU] * 3 + [1e-3] * 3
        else:
            axis_labels = [""] * 6
    elif axis_labels is None:
        axis_labels = [""] * 6

    if o.shape[dim_axis] == 2:
        subplot_cnt = (1, 2)
        subplot_perms = 2
    elif o.shape[dim_axis] == 3:
        subplot_cnt = (1, 3)
        subplot_perms = 3
    elif o.shape[dim_axis] == 4:
        subplot_cnt = (2, 3)
        subplot_perms = 6
    elif o.shape[dim_axis] == 5:
        subplot_cnt = (2, 5)
        subplot_perms = 10
    else:
        subplot_cnt = (3, 5)
        subplot_perms = 15
    subplot_cnt_ind = 1

    if "window" in options:
        size_in = options["window"]
        size_in = tuple(x / 80.0 for x in size_in)
    else:
        size_in = (19, 10)

    fig = plt.figure(figsize=size_in, dpi=80)

    fig.suptitle(
        options.get("title", "Orbital elements distribution"),
        fontsize=options.get("title_font_size", 24),
    )
    axes = []
    for I in range(subplot_perms):
        ax = fig.add_subplot(subplot_cnt[0], subplot_cnt[1], subplot_cnt_ind)
        axes.append(ax)

        if dim_axis == 1:
            x = o[:, axis_plot[I][0]] * scale[axis_plot[I][0]]
            y = o[:, axis_plot[I][1]] * scale[axis_plot[I][1]]
        else:
            x = o[axis_plot[I][0], :] * scale[axis_plot[I][0]]
            y = o[axis_plot[I][1], :] * scale[axis_plot[I][1]]

        sc = ax.scatter(
            x.flatten(),
            y.flatten(),
            marker=options.get("marker", "."),
            c=options.setdefault("color", "b"),
            s=options.get("size", 2),
        )

        if isinstance(options["color"], np.ndarray):
            plt.colorbar(sc)

        # x_ticks = np.linspace(np.min(o[:,axis_plot[I][0]]),np.max(o[:,axis_plot[I][0]]), num=4)
        # plt.xticks( [round(x,1) for x in x_ticks] )
        ax.set_xlabel(
            axis_labels[axis_plot[I][0]],
            fontsize=options.setdefault("ax_font_size", 22),
        )
        ax.set_ylabel(
            axis_labels[axis_plot[I][1]],
            fontsize=options["ax_font_size"],
        )
        plt.xticks(fontsize=options.setdefault("tick_font_size", 17))
        plt.yticks(fontsize=options["tick_font_size"])

        if limits is not None:
            if len(limits) > axis_plot[I][0]:
                ax.set_xlim(*limits[axis_plot[I][0]])
            if len(limits) > axis_plot[I][1]:
                ax.set_ylim(*limits[axis_plot[I][1]])

        subplot_cnt_ind += 1

    plt.tight_layout(rect=options.setdefault("tight_rect", [0, 0.03, 1, 0.95]))

    return fig, axes



from .space_object_population_table_plot import space_object_population_table_plot
from .schedule_plot import schedule_plot
from .azel_skyplot import azel_skyplot
from .ecef_states_positions_plot import ecef_states_positions_plot
from .ecef_states_positions_plot import ecef_states_positions_plot_letsplot
from .kepler_space_object_on_map import kepler_space_object_on_map
from .radar_schedule_ecef_position_plot import radar_schedule_ecef_position_plot



#!/usr/bin/env python
import pathlib
import sys
import time
import getpass
import json
import subprocess
import requests

from .commands import add_command


URL = "https://discosweb.esoc.esa.int"

_ID_MAP = {
    "NORAD": "satno",
    "COSPAR": "cosparId",
    "DISCOS": "id",
}


def main(args):
    if isinstance(args.object_id, str):
        args.object_id = [args.object_id]

    if len(args.output) == 0:
        output_pth = sys.stdout
    else:
        output_pth = pathlib.Path(args.output)
        if output_pth.exists() and not args.override:
            raise FileExistsError(f"{output_pth} exists")
        elif output_pth.exists() and args.override:
            print("Overriding existing file..")
            output_pth.unlink()
        output_pth = open(output_pth, "w")

    if args.type == "NORAD":
        tmp_oids = []
        for oid in args.object_id:
            oid = oid[:-1] if oid[-1].isalpha() else oid
            tmp_oids.append(oid)
        args.object_id = tmp_oids

    elif args.type == "COSPAR":
        args.object_id = [f"'{oid}'" for oid in args.object_id]

    if args.secret_tool_key is not None:
        res = subprocess.run(
            ["secret-tool", "lookup", "token"] + args.secret_tool_key,
            capture_output=True,
            text=True,
        )
        token = res.stdout
    elif args.credentials is not None:
        raise NotImplementedError("Add input of username/password from file")
    else:
        token = getpass.getpass("API token for:")

    current_page = 1
    params = {
        "sort": "id",
        "page[size]": 100,
    }
    if len(args.object_id) > 0:
        oids = ",".join(args.object_id)
        if len(args.object_id) == 1:
            filt_str = f"{_ID_MAP[args.type]}={oids}"
        else:
            filt_str = f"in({_ID_MAP[args.type]},({oids}))"
        params["filter"] = filt_str
        print(f'Fetching data for "{filt_str}"')
    else:
        print("Fetching all data")

    objs = []
    while True:
        print("getting page ", current_page)
        params["page[number]"] = current_page
        response = requests.get(
            f"{URL}/api/objects",
            headers={
                "Authorization": f"Bearer {token}",
                "DiscosWeb-Api-Version": "2",
            },
            params=params,
        )

        if response.status_code == 429:
            retry_interval = int(response.headers["Retry-After"])
            print(f"API requests exceeded, sleeping for {retry_interval} s")
            time.sleep(retry_interval + 1)
            continue
        else:
            current_page += 1

        doc = response.json()
        if response.ok:
            objs += doc["data"]
            print(f"{len(doc['data'])} Entries found...")
            if doc["links"]["next"] is None:
                print("No more pages, exiting...")
                break
        else:
            print("Error...")
            print(doc["errors"])

    print(f"{len(objs)} Entries downloaded...")
    json.dump(objs, output_pth, indent=2)
    if len(args.output) > 0:
        output_pth.close()


def parser_build(parser):
    parser.add_argument(
        "-t",
        "--type",
        type=str,
        default="NORAD",
        choices=["NORAD", "COSPAR", "DISCOS"],
        help='Input ID, ID variant can be set with "--type"',
    )
    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        default="",
        help="Path to file where output should be written.",
    )
    parser.add_argument(
        "--secret-tool-key",
        "-k",
        nargs=1,
        type=str,
        metavar="KEY",
        help='Attribute [named "token"] value [key] to fetch secret from',
    )
    parser.add_argument(
        "--credentials",
        "-c",
        nargs=1,
        metavar="FILE",
        help="File containing DISCOSweb token",
    )
    parser.add_argument(
        "--override",
        action="store_true",
        help="Override output file",
    )
    parser.add_argument(
        "object_id",
        metavar="ID",
        type=str,
        nargs="+",
        help='Input ID(s), ID variant can be set with "--type"',
    )

    return parser


add_command(
    name="discos",
    function=main,
    parser_build=parser_build,
    command_help="Download object information from ESA discosweb",
)



#!/usr/bin/env python

"""Miscellaneous functions"""

import numpy as np
import scipy.constants
import spacecoords
import pyorb


def equidistant_sampling(orbit, start_t, end_t, max_dpos=1e3, eccentricity_tol=0.3):
    """Find the temporal sampling of an orbit which is sufficient to achieve a
    maximum spatial separation. Assume elliptic orbit and uses Keplerian propagation
    to find sampling, does not take perturbation patterns into account. If
    eccentricity is small, uses periapsis speed and uniform sampling in time.

    :param pyorb.Orbit orbit: Orbit to find temporal sampling of.
    :param float start_t: Start time in seconds
    :param float end_t: End time in seconds
    :param float max_dpos: Maximum separation between evaluation points in meters.
    :param float eccentricity_tol: Minimum eccentricity below which the orbit is
        approximated as a circle and temporal samples are uniform in time.
    :return: Vector of sample times in seconds.
    :rtype: numpy.ndarray
    """
    if len(orbit) > 1:
        raise ValueError(f"Cannot use vectorized orbits: len(orbit) = {len(orbit)}")

    if orbit.e <= eccentricity_tol:
        r = pyorb.elliptic_radius(0.0, orbit.a, orbit.e, degrees=False)
        v = pyorb.orbital_speed(r, orbit.a, orbit.G * (orbit.M0 + orbit.m))[0]
        return np.arange(start_t, end_t, max_dpos / v)

    tmp_orb = orbit.copy()
    tmp_orb.auto_update = False

    tmp_orb.propagate(start_t)
    period = tmp_orb.period

    t_curr = start_t
    t = [t_curr]
    t_repeat = None
    while t_curr < end_t:
        if t_curr - start_t > period:
            if t_repeat is None:
                t_repeat = len(t)
            dt = t[-t_repeat + 1] - t[-t_repeat]
            t_curr += dt
        else:
            v = tmp_orb.speed[0]
            dt = max_dpos / v
            t_curr += dt
            tmp_orb.propagate(dt)

        t.append(t_curr)
    return np.array(t, dtype=np.float64)


def calculate_range(enu):
    """Norm of the ENU coordinates."""
    return np.linalg.norm(enu[:3, :], axis=0)


def calculate_range_rate(enu):
    """Projected ENU velocity along the ENU range."""
    return np.sum(enu[3:, :] * (enu[:3, :] / np.linalg.norm(enu[:3, :], axis=0)), axis=0)


def calculate_zenith_angle(enu, radians=False):
    """Zenith angle of the ENU coordinates."""
    return spacecoords.linalg.vector_angle(
        np.array([0, 0, 1], dtype=np.float64), enu[:3, :], degrees=not radians
    )


def signal_delay(st1, st2, ecef):
    """Signal delay due to speed of light between station-1 to ecef position to station-2"""
    r1 = np.linalg.norm(ecef - st1.ecef[:, None], axis=0)
    r2 = np.linalg.norm(ecef - st1.ecef[:, None], axis=0)
    dt = (r1 + r2) / scipy.constants.c
    return dt


def instantaneous_to_coherrent(gain, groups, N_IPP, IPP_scale=1.0, units="dB"):
    """Using pulse encoding schema, subgroup setup and coherent integration setup; convert from instantaneous gain to coherently integrated gain.

    :param float gain: Instantaneous gain, linear units or in dB.
    :param int groups: Number of subgroups from witch signals are coherently combined, assumes subgroups are identical.
    :param int N_IPP: Number of pulses to coherently integrate.
    :param float IPP_scale: Scale the IPP effective length in case e.g. the IPP is the same but the actual TX length is lowered.
    :param str units: If string equals 'dB', assume input and output units should be dB, else use linear scale.

    :return float: Gain after coherent integration, linear units or in dB.
    """
    if units == "dB":
        return gain + 10.0 * np.log10(groups * N_IPP * IPP_scale)
    else:
        return gain * (groups * N_IPP * IPP_scale)


def coherrent_to_instantaneous(gain, groups, N_IPP, IPP_scale=1.0, units="dB"):
    """Using pulse encoding schema, subgroup setup and coherent integration setup; convert from coherently integrated gain to instantaneous gain.

    :param float gain: Coherently integrated gain, linear units or in dB.
    :param int groups: Number of subgroups from witch signals are coherently combined, assumes subgroups are identical.
    :param int N_IPP: Number of pulses to coherently integrate.
    :param float IPP_scale: Scale the IPP effective length in case e.g. the IPP is the same but the actual TX length is lowered.
    :param str units: If string equals 'dB', assume input and output units should be dB, else use linear scale.

    :return float: Instantaneous gain, linear units or in dB.
    """
    if units == "dB":
        return gain - 10.0 * np.log10(groups * N_IPP * IPP_scale)
    else:
        return gain / (groups * N_IPP * IPP_scale)



#!/usr/bin/env python

"""Errors

"""
# Python standard import
from abc import ABC, abstractmethod
import types
import pathlib

# Third party import
import numpy as np


# Local import


class Errors(ABC):
    """A standardized way for adding random errors to data

    The methods corresponding to the variables should take the data and return
    """

    VARIABLES = []

    def __init__(self, seed=None):
        self.seed = seed
        self._check_methods()

    def set_numpy_seed(self):
        """This should be called before any error generating methods if reproduction of results needs to be ensured."""
        if self.seed is not None:
            self.seed += 1
            np.random.seed(self.seed)

    def get(self, var, data, *args, **kwargs):
        func = getattr(self, var)
        return func(data, *args, **kwargs)

    def _check_method(self, name):
        return hasattr(self, name) and type(getattr(self, name)) == types.MethodType

    def _check_methods(self):
        for var in self.VARIABLES:
            if not self._check_method(var):
                raise AttributeError(
                    f"Could not find implemented error generating method for {var}"
                )



#!/usr/bin/env python

"""wrapper for the SGP4 propagator"""

import logging
from dataclasses import dataclass
import numpy as np
from astropy.time import TimeDelta, Time
from .base import Propagator
from sorts.types import Settings, Frames, NDArray_N, NDArray_6xN
from sorts.space_object import SpaceObject
from sorts.utils import convert_to_relative_time
import spacecoords.celestial as cel

logger = logging.getLogger(__name__)


@dataclass
class KeplerSettings(Settings):
    numerical_tolerance: float = 1e-12
    max_iterations: int = 5000
    laguerre_degree: int = 5
    out_frame: Frames = "GCRS"
    internal_frame: Frames = "GCRS"


class Kepler(Propagator[KeplerSettings]):
    """Propagator class implementing the Kepler propagator,
    the propagation always occurs in GCRS frame.

    Frame options are found in the `sorts.frames.convert` function.

    """

    def propagate(
        self,
        space_object: SpaceObject,
        times: Time | TimeDelta | NDArray_N,
    ) -> NDArray_6xN:
        logger.debug("Kepler:propagate")
        tv = convert_to_relative_time(space_object.epoch, times)

        orb = space_object.state.copy()
        if space_object.frame != self.settings.internal_frame:
            orb._cart = cel.convert(
                space_object.epoch,
                orb._cart,
                in_frame=space_object.frame,
                out_frame=self.settings.internal_frame,
                frame_kwargs={},
            )
            orb.calculate_kepler()

        orb.direct_update = False
        orb.auto_update = False
        orb.solver_options = dict(
            tol=self.settings.numerical_tolerance,
            max_iter=self.settings.max_iterations,
            degree=self.settings.laguerre_degree,
        )

        orb.add(num=len(tv) - 1)
        kep0 = orb._kep[:, 0]
        orb._kep[:, :] = kep0[:, None]
        orb.propagate(tv)
        orb.calculate_cartesian()

        if self.settings.out_frame != self.settings.internal_frame:
            orb._cart = cel.convert(
                space_object.epoch + TimeDelta(tv, format="sec"),
                orb._cart,
                in_frame=self.settings.internal_frame,
                out_frame=self.settings.out_frame,
                frame_kwargs={},
            )

        logger.debug("Kepler:propagate:completed")

        return orb._cart



from __future__ import annotations
import logging, typing as t
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from sorts.const import max_datetime64_us, min_datetime64_us
from sorts import utils
from . import schedule
from .schedule import Schedule


logger = logging.getLogger(__name__)


_SK = schedule._K

DsIntermediateVarKey = t.Literal[
    "cummax_start_time",
    "cummax_end_time",
    "is_overlaped",
]


class _IK:
    """Internal helper class for accessing string keys consistently"""

    cummax_start_time: t.Final = "cummax_start_time"
    cummax_end_time: t.Final = "cummax_end_time"
    is_overlaped: t.Final = "is_overlaped"


utils.assert_class_attributes_equal_to(_IK, t.get_args(DsIntermediateVarKey))


def _inject_intermediate_columns(sch: Schedule) -> Schedule:
    sch[_IK.cummax_start_time] = xr.full_like(
        sch[_SK.multi_index], np.datetime64("NaT"), dtype="datetime64[us]"
    )
    sch[_IK.cummax_end_time] = xr.full_like(
        sch[_SK.multi_index], np.datetime64("NaT"), dtype="datetime64[us]"
    )
    sch[_IK.is_overlaped] = xr.full_like(sch[_SK.multi_index], False, dtype=np.bool)

    return sch


def _propagate_cummax_start_time_cummax_end_time(merged_sch: Schedule) -> Schedule:
    # todo: why is this needed? need some more logic explanation
    merged_sch[_IK.cummax_start_time] = (
        t.cast(pd.Series, merged_sch[_IK.cummax_start_time].to_pandas())
        .groupby(level=[_SK.stn_num])
        .ffill()
    )
    merged_sch[_IK.cummax_end_time] = (
        t.cast(pd.Series, merged_sch[_IK.cummax_end_time].to_pandas())
        .groupby(level=[_SK.stn_num])
        .ffill()
    )

    return merged_sch


def _remove_entries_with_time_clash(merged_sch_: Schedule, incoming_sch: Schedule) -> Schedule:
    # NOTE:
    #   - the `allowed_start_time` has the previous `end_time`;
    #     for top values that have no corresponding pervious values, `min_datetime64_us` is used
    #   - the `allowed_end_time` has the next `start_time`;
    #     for bottom values that have no corresponding next values, `max_datetime64_us` is used
    allowed_start_time = (
        t.cast(pd.Series, merged_sch_[_IK.cummax_end_time].to_pandas())
        .groupby(level=_SK.stn_num)
        .shift(+1, fill_value=min_datetime64_us)
    )
    allowed_end_time = (
        t.cast(pd.Series, merged_sch_[_IK.cummax_start_time].to_pandas())
        .groupby(level=_SK.stn_num)
        .shift(-1, fill_value=max_datetime64_us)
    )

    # calc bool mask for overlapping just for the new/incoming entries
    is_over_allowed_start_time = merged_sch_[_SK.start_time].loc[
        {_SK.multi_index: incoming_sch[_SK.multi_index]}
    ] < allowed_start_time.reindex(incoming_sch[_SK.multi_index].to_pandas().index)

    is_over_allowed_end_time = merged_sch_[_SK.end_time].loc[
        {_SK.multi_index: incoming_sch[_SK.multi_index]}
    ] > allowed_end_time.reindex(incoming_sch[_SK.multi_index].to_pandas().index)

    # NOTE: we checked for is_overlaped instead of is_allowed
    #   so that it is safe agaisnt comparison with `NaT`, which always return false
    #   (and we assume `NaT` mean "no restructions" for both allowed_start_time and allowed_end_time)
    merged_sch_[_IK.is_overlaped].loc[{_SK.multi_index: incoming_sch[_SK.multi_index]}] = (
        is_over_allowed_start_time
    ) | is_over_allowed_end_time

    merged_sch_ = merged_sch_.loc[{_SK.multi_index: ~merged_sch_[_IK.is_overlaped]}]

    return merged_sch_


def _remove_entries_without_corresponding_tx(
    merged_sch: Schedule,
    incoming_sch: Schedule,
    exp_id_stn_id_pairs_map: schedule.ExperimentIdStationIdPairsMap,
) -> Schedule:
    exp_ids: npt.NDArray[np.int16] = np.unique(incoming_sch[_SK.exp_num].to_numpy())
    for exp_id in exp_ids:
        stn_pairs = exp_id_stn_id_pairs_map[exp_id]

        for tx_stn_num, rx_stn_num in stn_pairs:
            incoming_tx_entries = t.cast(
                pd.Series,
                incoming_sch.loc[{_SK.multi_index: (exp_id, tx_stn_num, slice(None), slice(None))}][
                    _SK.multi_index
                ].to_pandas(),
            )

            is_tx_dropped = ~incoming_tx_entries.isin(merged_sch[_SK.multi_index].to_pandas())
            dropped_tx_start_time_idx = (
                incoming_sch.loc[{_SK.multi_index: is_tx_dropped[is_tx_dropped == True].index}]
                .indexes[_SK.multi_index]
                .get_level_values(_SK.start_time)
            )

            # NOTE: we cannot filter the MultiIndex by levels 'start_time', 'exp_num', 'stn_num'
            #   directly using `(dropped_tx_start_time_idx, exp_id, rx_stn_num)` here
            #   because 'dropped_tx_start_time_idx' can contains entries from other exp or station, which will lead to KeyError.
            #
            #   Instead, we filter by first by levels 'exp_num', 'stn_num' here
            #   and followed by an intersection with 'dropped_tx_start_time_idx' later when dropping
            try:
                rx_start_time_of_the_exp_idx = (
                    merged_sch.loc[
                        {
                            _SK.multi_index: (
                                exp_id,
                                rx_stn_num,
                                slice(None),
                                slice(None),
                            )
                        }
                    ]
                    .indexes[_SK.multi_index]
                    .get_level_values(_SK.start_time)
                )
            except KeyError:
                # NOTE: it is possible that there is no entrise with multi_index = (ANY, exp_id, rx_stn_num, ANY)
                #   we can skip the dropping such case
                rx_start_time_of_the_exp_idx = None

            if rx_start_time_of_the_exp_idx is not None:
                # NOTE: cannot drop using `.drop_sel` directly for some reason, so we do a selection by `.loc` first
                dropping = merged_sch.loc[
                    {
                        _SK.multi_index: (
                            exp_id,
                            rx_stn_num,
                            slice(None),
                            rx_start_time_of_the_exp_idx.intersection(dropped_tx_start_time_idx),
                        )
                    }
                ]
                merged_sch = merged_sch.drop_sel({_SK.multi_index: dropping[_SK.multi_index]})
            else:
                continue

    return merged_sch


def _update_cummax_start_time_cummax_end_time(merged_sch: Schedule) -> Schedule:
    merged_sch[_IK.cummax_start_time] = (
        t.cast(pd.Series, merged_sch[_SK.start_time].to_pandas())
        .groupby(level=[_SK.stn_num])
        .cummax()
    )

    merged_sch[_IK.cummax_end_time] = (
        t.cast(pd.Series, merged_sch[_SK.end_time].to_pandas())
        .groupby(level=[_SK.stn_num])
        .cummax()
    )

    return merged_sch


def priority_scheduling(
    schs: t.Sequence[Schedule], exp_id_stn_id_pairs_map: schedule.ExperimentIdStationIdPairsMap
) -> Schedule:
    """
    Merge a sequence of schedule data for a single station into one,
    schedule with smaller index in the sequence is given priority over those with larger index.

    Note: It is assumed (and not checked) that each of the schedule itself does not contain overlapping entries.

    Known artifacts:
      There could be cases of dangling TX entries in the result
      (but not the inverse, i.e. there will not be any dangling RX entries).
    """

    # The logic of this function:
    # 1. prepare an empty schedule data as the merge result
    # 2. for each of the schedule passed in
    #   2.1. merge it into the merge result
    #   2.2. propagate `cummax_start_time`, `cummax_end_time` to new rows
    #   2.3. filter out new rows that have conflicts
    #   2.3. update `cummax_start_time`, `cummax_end_time`

    # early return for empty case
    if len(schs) == 0:
        logger.warning(
            f"An empty `Schedule` is being generated by `{priority_scheduling.__name__}`."
            + "This happens when an empty `Schedule` list is being passed as argument, which is usually not intended."
        )
        return schedule.empty()

    # init an empty dataset for a schedule and add some columns, will be used store merged schedule
    merged_sch = schedule.empty()
    if len(schs) > 0:
        merged_sch.attrs = schs[0].attrs
    merged_sch = _inject_intermediate_columns(merged_sch)
    for incoming_sch in schs:
        incoming_sch = _inject_intermediate_columns(incoming_sch)

        # merge and then sort the schedule
        # we use a "stable" sorting algo to retains relative order,
        # so the df will be in order of start_time, then priority after sorting
        # also note that attrs merged in the way that former schedule has higher priority than latter,
        # consistent with the func `priority_scheduling`
        merged_sch = xr.concat([merged_sch, incoming_sch], dim=_SK.multi_index)
        merged_sch = merged_sch.sortby(_SK.start_time)

        merged_sch = _propagate_cummax_start_time_cummax_end_time(merged_sch)
        merged_sch = _remove_entries_with_time_clash(merged_sch, incoming_sch)
        merged_sch = _remove_entries_without_corresponding_tx(
            merged_sch, incoming_sch, exp_id_stn_id_pairs_map
        )
        merged_sch = _update_cummax_start_time_cummax_end_time(merged_sch)

    merged_sch = merged_sch.drop_vars(
        [
            _IK.cummax_start_time,
            _IK.cummax_end_time,
            _IK.is_overlaped,
        ]
    )

    return merged_sch



"""rapper for the SGP4 propagator

"""

# Python standard import
import logging
from copy import copy

# Third party import
import numpy as np
import poliastro
from poliastro.twobody.propagation import farnocchia
from poliastro.bodies import Earth
import astropy.units as units


# Local import
from .base import Propagator
from .. import frames

logger = logging.getLogger(__name__)

class TwoBody(Propagator):
    """Propagator class implementing the Kepler propagator using `poliastro`,
    the propagation always occurs in GCRS frame.

    Frame options are found in the `sorts.frames.convert` function.

    :ivar str in_frame: String identifying the input frame.
    :ivar str out_frame: String identifying the output frame.

    :param str in_frame: String identifying the input frame.
    :param str out_frame: String identifying the output frame.
    """

    DEFAULT_SETTINGS = copy(Propagator.DEFAULT_SETTINGS)
    DEFAULT_SETTINGS.update(
        dict(
            out_frame="GCRS",
            in_frame="GCRS",
            method=farnocchia,
            body=Earth,
        )
    )

    def __init__(self, settings=None, **kwargs):
        super(TwoBody, self).__init__(settings=settings, **kwargs)
        logger.debug("sorts.propagator.TwoBody:init")

    def propagate(self, t, state0, epoch, **kwargs):
        """Propagate a state

        :param float/list/numpy.ndarray/astropy.time.TimeDelta t: Time to
            propagate relative the initial state epoch.
        :param float/astropy.time.Time epoch: The epoch of the initial state.
        :param numpy.ndarray state0: 6-D Cartesian state vector in SI-units.
        :param bool radians: If true, all angles are assumed to be in radians.
        :return: 6-D Cartesian state vectors in SI-units.

        """
        logger.debug(f"TwoBody:propagate:len(t) = {len(t)}")

        t, epoch = self.convert_time(t, epoch)
        times = epoch + t

        state0_GCRS = frames.convert(
            epoch,
            state0,
            in_frame=self.settings["in_frame"],
            out_frame="GCRS",
        )

        poli_orb = poliastro.twobody.Orbit.from_vectors(
            self.settings["body"],
            state0_GCRS[:3] * units.m,
            state0_GCRS[3:] * units.m / units.s,
            epoch=epoch,
        )
        astropy_state_GCRS = poliastro.twobody.propagation.propagate(
            poli_orb,
            t.sec * units.s,
            method=self.settings["method"],
            **kwargs,
        )
        states_GCRS = np.empty((6, len(t)), dtype=np.float64)
        states_GCRS[:3, :] = astropy_state_GCRS.xyz.to(units.m).value
        states_GCRS[3:, :] = astropy_state_GCRS.differentials["s"].d_xyz.to(units.m / units.s).value

        states = frames.convert(
            times,
            states_GCRS,
            in_frame="GCRS",
            out_frame=self.settings["out_frame"],
        )

        logger.debug("TwoBody:propagate:completed")

        return states



import numpy as np
import scipy.constants
import pyant

from ..signals import rayleigh_separatrix_range, hard_target_diameter

T_REF = 290  # K


def estimate_radar_parameters(
    frequency,
    antenna_num,
    antenna_spacing_lambda,
    antenna_efficiency,
    antenna_input_power,
    thermal_load,
    noise_figure_db,
    amplifier_gain_db,
    insertion_loss_db,
    duty_cycle=0.1,
    computation_power_draw_scaling=2,
    t_sky=30,
    reference_snr_db=10.0,
    reference_ranges=1000e3,
    coherent_integration_time=0.02,
    bandwidth_limit_ratio=10,
):
    """
    todo

    ref: kildal book, other books
    """
    lam = scipy.constants.c / frequency

    tx_power = antenna_num * antenna_input_power

    thermal_production = antenna_num * thermal_load * duty_cycle
    radar_power_draw = antenna_num * (thermal_load + antenna_input_power) * duty_cycle
    total_power_draw = computation_power_draw_scaling * radar_power_draw
    # avg antenna package size
    size = antenna_spacing_lambda * lam

    # approx the circle radius of all antenna package area
    # A_tot = size ** 2 * antenna_num
    # r = np.sqrt(A_tot/np.pi) = np.sqrt(antenna_num / np.pi) * size
    radius = np.sqrt(antenna_num / np.pi) * size

    # just based on the aperture
    directivity = 4 * (np.pi * radius) ** 2 / lam**2

    # the peak gain, this is an approximation in the main lobe since
    # the gain will no longer be "directive gain" when divided by
    # the antenna_efficiency as the antenna gain would also modify
    # the side lobes and such
    peak_gain = antenna_efficiency * directivity

    beam = pyant.models.Gaussian(
        peak_gain=peak_gain,
    )
    param = pyant.models.GaussianParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        normal_pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=frequency,
        radius=radius,
        beam_width_scaling=1.22,
    )
    half_power_angle = np.degrees(param.beam_width_scaling * lam / (radius * 2))

    # the radar equation will give us the power induced in the system after antenna reception, now
    # to turn this into a SNR, we can assume that this power stays constant and instead just modify
    # the noise power accordingly as a function of all the effects of the system

    # assume we have a noise of T_0 at the digitisers
    # doing mean over the different antenna signals will keep T_0 constant but increase signal
    # strength in a coherent fashion, this is covered by the gain function (usually done trough the
    # "array factor"), hence we can keep T_0 constant trough the lossless combiner.

    # this noise is pretty much only the LNA from the system side and the sky noise from the
    # external side, the noise is additive so the following equation should suffice

    # The noise factor is the ratio of input SNR to output SNR, i.e. the noise introduced by the
    # component. The noise figure is the decibel equivalent version of the noise factor.
    #
    # F = 1 + T_noise / T_ref
    # F = SNR_in / SNR_out
    amplifier_gain = 10 ** (amplifier_gain_db / 10)
    noise_figure = 10 ** (noise_figure_db / 10)
    t_noise = amplifier_gain * t_sky + (noise_figure - 1) * T_REF
    t_noise = t_noise / bandwidth_limit_ratio

    # Then coherent integration will happen, and here, whatever our sampling frequency is gets
    # removed by the fact that we are coherently integrating, and all that matters is the coherent
    # integration time and the ratio of bandwidth reduction by any supposed filtering
    rayleigh_separatrix = rayleigh_separatrix_range(
        beam.peak_gain,
        beam.peak_gain,
        lam,
        tx_power,
        10 ** (reference_snr_db / 10),
        coherent_integration_time=coherent_integration_time,
        effective_noise_temperature=t_noise,
        radar_albedo=1.0,
    )
    effective_bandwidth = 1 / coherent_integration_time
    minimum_diameter = [
        hard_target_diameter(
            beam.peak_gain,
            beam.peak_gain,
            lam,
            tx_power,
            reference_range,
            reference_range,
            10 ** (reference_snr_db / 10),
            bandwidth=effective_bandwidth,
            rx_noise_temp=t_noise,
            radar_albedo=1.0,
        )
        for reference_range in reference_ranges
    ]
    data = dict(
        rayleigh_separatrix=rayleigh_separatrix,
        minimum_diameter=minimum_diameter,
        t_noise=t_noise,
        tx_power=tx_power,
        thermal_production=thermal_production,
        radar_power_draw=radar_power_draw,
        total_power_draw=total_power_draw,
        beam=beam,
        beam_parameters=param,
        half_power_angle=half_power_angle,
        effective_bandwidth=effective_bandwidth,
    )
    return data
